
class SecureToken:
    def __init__(self):
        self.email = None
        self.access_token = None
        self.refresh_token = None
        self.access_token_expires_at = None
        self.refresh_token_expires_at = None




import json

from qb_core.common.base_model import BaseModel


class Customer(BaseModel):

    def __init__(self):
        self.name = ""
        self.email = ""
        self.phone_number = ""
        self.address = ""
        self.card_info = ""

    def init(self, **kwargs):
        for k, v in kwargs.items():
            if k == "name":
                self.name = v
            if k == "email":
                self.email = v
            if k == "phone_number":
                self.phone_number = v
            if k == "address":
                self.address = v
            if k == "card_info":
                self.card_info = v

    def __str__(self):
        return f'{self.name=}, {self.email=}'


class BaseCustomerEvent:
    def __init__(self, customer):
        self.customer = customer

    def __repr__(self):
        return json.dumps(self.__dict__, default=str)

    def __str__(self):
        return f'{self.customer=}'


class CustomerAdded(BaseCustomerEvent):
    pass


class CustomerRemoved(BaseCustomerEvent):
    pass





# this interface defines all the possible access patterns and write
# patterns for the cook object
# the plugin that implements this interface can use this information
# to define the appropriate storage and efficient ways to provide the
# handle data with that storage
class CookPersistenceInterface:
    def __init__(self):
        pass

    def read_cook_count(self):
        pass

    def read_cook(self, cook_name):
        pass

    def save_cook(self, cook):
        pass

    def remove_cook(self, cook):
        pass

    def print_all(self):
        pass



import json

from qb_core.common.base_model import BaseModel


class OrderItem(BaseModel):
    def __init__(self):
        self.item_name = ""
        self.quantity = 0
        self.price = 0.0
        self.special_instruction = ""

    def init(self, **kwargs):
        for k, v in kwargs.items():
            if k == "item_name":
                self.item_name = v
            if k == "quantity":
                self.quantity = v
            if k == "price":
                self.price = v
            if k == "special_instruction":
                self.special_instruction = v

    def __str__(self):
        return f'{self.item_name=}, {self.quantity=}'


class CookAssignment(BaseModel):
    def __init__(self):
        self.cook_name = ""
        self.cook_email = ""
        self.time_cook_picked_up = ""
        self.time_cook_completed = ""

    def init(self, **kwargs):
        for k, v in kwargs.items():
            if k == "cook_name":
                self.cook_name = v
            if k == "cook_email":
                self.cook_email = v
            if k == "time_cook_picked_up":
                self.time_cook_picked_up = v
            if k == "time_cook_completed":
                self.time_cook_completed = v

    def __str__(self):
        return f'cook_name={self.cook_name}, ' \
               f'cook_picked_up={self.time_cook_picked_up != ""}, ' \
               f'cook_completed={self.time_cook_completed != ""}'


class Order(BaseModel):
    def __init__(self):
        self.id = ""
        self.customer_email = ""
        self.total_price = 0.0
        self.items = []
        self.cook_assignment = None

    def init(self, **kwargs):
        for k, v in kwargs.items():
            if k == "total_price":
                self.total_price = v
            if k == "items":
                self.items = v
            if k == "id":
                self.id = v

    def __str__(self):
        return f'customer_email={self.customer_email}, ' \
               f'total_price={self.total_price}, ' \
               f'number_of_items={len(self.items)} ' \
               f'cook_assignment=({str(self.cook_assignment)})'


class BaseOrderEvent:
    def __init__(self, order):
        self.order = order

    def __repr__(self):
        return json.dumps(self.__dict__, default=str)

    def __str__(self):
        return f'{self.order=}'


class OrderReadyForPickup(BaseOrderEvent):
    pass


class OrderInProgress(BaseOrderEvent):
    pass



import json

from qb_core.common.base_model import BaseModel


class Cook(BaseModel):

    def __init__(self):
        self.name = ""
        self.email = ""

    def init(self, **kwargs):
        for k, v in kwargs.items():
            if k == "name":
                self.name = v
            if k == "email":
                self.email = v

    def __str__(self):
        return f'{self.name=}, {self.email=}'


class BaseCookEvent:
    def __init__(self, cook):
        self.cook = cook

    def __repr__(self):
        return json.dumps(self.__dict__, default=str)

    def __str__(self):
        return f'{self.cook=}'


class CookAdded(BaseCookEvent):
    pass


class CookRemoved(BaseCookEvent):
    pass



import pytest

from qb_core.rats_pkg.cook import cook_persistence_core
from qb_core.rats_pkg.cook.cook_model import Cook, CookAdded, CookRemoved
from qb_core.rats_pkg.restaurant import _restaurant_persistence
from qb_core.rats_pkg.restaurant.restaurant_model import Restaurant


def test_restaurant_model():
    restaurant = Restaurant()

    restaurant.init(name="n", address="a",
                    cuisine_type="ct", rating="r",
                    status="s", manager_name="mn",
                    manager_email="me")
    assert str(restaurant) == "self.name='n', self.manager_name='mn', me"
    assert repr(restaurant) == '{"name": "n", "address": "a", ' \
                               '"cuisine_type": "ct", "rating": "r", ' \
                               '"status": "s", "manager_name": "mn", ' \
                               '"manager_email": "me", "verification_code": null, ' \
                               '"verification_valid_till": null, "cook_dict": {}}'


# is this needed?
# just trigger something to get 100% coverage.
# i am not sure if this is needed
# it will be nice if i can roll it with main functioal test
def test_save_self_verification_code():
    restaurant = Restaurant()
    restaurant.save_verification_code("1234")
    assert restaurant.verification_code == "1234"
    assert restaurant.status == "pending_verification"


@pytest.fixture()
def setup_restaurant_persistence():
    # making sure the test always starts with empty list
    _restaurant_persistence.RESTAURANT_DICT = {}
    yield
    # making sure the test cleans up to an empty list
    _restaurant_persistence.RESTAURANT_DICT = {}


def test_restaurant_persistence(setup_restaurant_persistence):
    assert _restaurant_persistence.read_restaurant_count() == 0

    expected_restaurant = Restaurant()
    expected_restaurant.init(name="r1", manager_email="r1me@unreal.com")

    _restaurant_persistence.save_restaurant(expected_restaurant)
    actual_restaurant = _restaurant_persistence.read_restaurant_by_manager_email("r1me@unreal.com")
    assert actual_restaurant == expected_restaurant

    actual_restaurant = _restaurant_persistence.read_restaurant_by_manager_email("some_junk")
    assert actual_restaurant is None

    _restaurant_persistence.print_all()



"""
use this module to initialize the whole app

Note setuptools is not picking this up
probably because this module is directly off of src
"""
import qb_core.oproc_pkg.oproc
import qb_core.rats_pkg.rats
import qb_core.ocra_pkg.ocra
from qb_core.common.plugin.plugin_manager import PluginManager


def initialize_app():
    # since this involves importing, it's good to do this upfront
    # PluginManager.setup_plugins("cook_persistence_plug_point", "qb_core.rats_pkg.cook.cook_persistence_core")
    # empty override dictionary implies just use core plugins
    # PluginManager.override_plugins({})
    # print(f'{PluginManager.get_plugin("cook_persistence_plug_point")}')

    qb_core.rats_pkg.rats.initialize_module()
    qb_core.oproc_pkg.oproc.initialize_module()
    qb_core.ocra_pkg.ocra.initialize_module()





"""
for now, define this as a simple module. we could later make this into a class if necessary

this module has everything needed to represent Restaurant entity

attributes:
    name
    address
    cuisine_type
    rating
    status
"""

from qb_core.common.base_model import BaseModel


class Restaurant(BaseModel):

    def __init__(self):
        self.name = ""
        self.address = ""
        self.cuisine_type = ""
        self.rating = ""
        self.status = ""
        # for now, implement manager info right here
        self.manager_name = ""
        self.manager_email = ""
        self.verification_code = None
        self.verification_valid_till = None
        self.cook_dict = {}

    def init(self, **kwargs):
        for k, v in kwargs.items():
            if k == "name":
                self.name = v
            if k == "address":
                self.address = v
            if k == "cuisine_type":
                self.cuisine_type = v
            if k == "rating":
                self.rating = v
            if k == "status":
                self.status = v
            if k == "manager_name":
                self.manager_name = v
            if k == "manager_email":
                self.manager_email = v

    def __str__(self):
        return f'{self.name=}, {self.manager_name=}, {self.manager_email}'

    def save_verification_code(self, code):
        self.verification_code = code
        # NICE_TODO Feature - Security - set this to 2 days
        self.verification_valid_till = None
        self.status = "pending_verification"



COOK_DICT = {}


def read_cook_count():
    return len(COOK_DICT)


def read_cook(cook_name):
    return COOK_DICT.get(cook_name)


def save_cook(cook):
    COOK_DICT[cook.name] = cook


def remove_cook(cook):
    COOK_DICT.pop(cook.name)


def print_all():
    print(f'{COOK_DICT=}')



core_rats_plug_point_configuration = {
    "cook_persistence_plug_point": {
        "plugin": "qb_core.rats_pkg.cook.cook_persistence_core",
        "description": "provides a core persistence for cook entity",
        "pluggable": "cook_entity"  # python module that uses the plug point
    },
    # things to be added in the future
    # "menu_persistence_plug_point": {
    #     "plug_point": "menu_persistence",
    #     "plugin": "core_menu_persistence",
    #     "description": "provides a core persistence for menu entity",
    #     "pluggable": "menu_processor"
    # },
    # "cook_added_event_plug_point": {
    #     "plugin": "core_cook_added_event",
    #     "description": "provides a core persistence for menu entity",
    #     "pluggable": "cook_processor"
    # }
}



# events emitted by RATS
EVENT_COOK_ADDED = "cook_added"
EVENT_RESTAURANT_ADDED = "restaurant_added"
EVENT_MENU_UPLOADED = "menu_uploaded"

# events emitted by OCRA
EVENT_ORDER_CONFIRMED = "order_confirmed"
EVENT_RESTAURANT_RATED = "restaurant_rated"
EVENT_DRIVER_RATED = "driver_rated"

# events emitted by OPROC
EVENT_ORDER_READY_FOR_PICKUP = "order_ready_for_pickup"
EVENT_ORDER_IN_PROGRESS = "order_in_progress"



from qb_core.ocra_pkg.order import _order_persistence


def read_orders(customer_email):
    return _order_persistence.read_orders(customer_email)


def save_order(order):
    print(f'OCRA save_order')
    _order_persistence.save_order(order)



"""
this tests cases and conditions that business scenarios cannot cover easily
ideal for odd error cases, edge cases
"""
import pytest

from qb_core.common.plugin.plugin_manager import PluginManager
from qb_core.rats_pkg.cook.cook_model import Cook, CookAdded, CookRemoved
from qb_core.rats_pkg.cook.cook_persistence_plug_point import CookPersistenceInterface

_cook_persistence = None


@pytest.fixture()
def setup_cook_persistence():
    # making sure the test always starts with empty list
    global _cook_persistence
    _cook_persistence = PluginManager.get_plugin("cook_persistence_plug_point")
    _cook_persistence.clear_all()

    yield
    # making sure the test cleans up to an empty list
    _cook_persistence.clear_all()


def test_cook_model():
    """this test is for testing str and repr methods"""
    cook = Cook()

    cook.init(name="nala paaka", email="np@unreal.com")
    assert str(cook) == "self.name='nala paaka', self.email='np@unreal.com'"
    assert repr(cook) == '{"name": "nala paaka", "email": "np@unreal.com"}'

    cook_removed = CookRemoved(cook)
    assert str(cook_removed) == 'self.cook={"name": "nala paaka", "email": "np@unreal.com"}'
    assert repr(cook_removed) == '{"cook": "self.name=\'nala paaka\', self.email=\'np@unreal.com\'"}'


def test_cook_persistence(setup_cook_persistence):
    """this test is for testing methods not commonly used"""
    assert _cook_persistence.read_cook_count() == 0

    expected_cook = Cook()
    expected_cook.init(name="nala paaka", email="np@unreal.com")

    actual_cook = _cook_persistence.read_cook("nala paaka")
    assert actual_cook is None

    _cook_persistence.save_cook(expected_cook)
    actual_cook = _cook_persistence.read_cook("nala paaka")
    assert actual_cook == expected_cook

    _cook_persistence.print_all()


def test_cook_persistence_plug_point():
    """since this is like an abstract class, business scenarios will not reach this code"""
    obj = CookPersistenceInterface()
    assert obj.read_cook_count() is None
    assert obj.read_cook("dont_care") is None
    assert obj.save_cook("dont_care") is None
    assert obj.remove_cook("dont_care") is None
    assert obj.print_all() is None



import json

from qb_core.rats_pkg.restaurant.restaurant_model import Restaurant

RESTAURANT_DICT = {}


def read_restaurant_count():
    return len(RESTAURANT_DICT)


def read_restaurant(restaurant_name):
    return RESTAURANT_DICT.get(restaurant_name)


def read_restaurant_by_manager_email(manager_email):
    # for now do this the hard way, loop through data.
    # when you write actual persistence you can index this appropriately
    for restaurant in RESTAURANT_DICT.values():
        if restaurant.manager_email == manager_email:
            return restaurant
    return None


def save_restaurant(restaurant):
    RESTAURANT_DICT[restaurant.name] = restaurant


def print_all():
    print(f'{json.dumps(RESTAURANT_DICT, default=str)=}')




"""
test_business_scenarios.py
use this module to test everything. not just unit tests but the whole app.
Why is this good
 * the whole app can run locally in a single process

TODO
 * Make sure each test is independent of the other.
"""
import asyncio
from datetime import datetime

import pytest

import core_main
from qb_core.common.plugin.plugin_manager import PluginManager
from qb_core.event_bus.event_bus_interface import EventBusInterface
from qb_core.ocra_pkg import ocra
from qb_core.oproc_pkg import oproc
from qb_core.rats_pkg import rats
from qb_core.rats_pkg.cook.cook_persistence_core import CookPersistenceCore
from qb_core.rats_pkg.cook import cook_persistence_core
from qb_core.rats_pkg.restaurant import _restaurant_persistence
from qb_core.rats_pkg.signin import _signin_persistence
from qb_core.rats_pkg.signup import _signup_persistence

print("lets test as we code")

"""
* this is the period the main thread pauses, typically after performing steps that
trigger an event. 
* by sleeping for this wait period we give asyncio opportunity to kick of listener
and the listner to complete it's steps.
* since we do in memory options in core, listener should take very short time 
* if listeners take longer increase this. 

* core event bus uses asyncio. it looks like ayncio works on 
one thread at a time (TODO is this right? research it, say print thread info)  
"""
PAUSE_PERIOD = 0.1

# _cook_persistence = CookPersistenceCore()
_cook_persistence = None


# NOTE: this test needs to be the first test as it initializes the app
@pytest.mark.asyncio
async def test_signup_happy_path_italian():
    core_main.initialize_app()
    global _cook_persistence
    _cook_persistence = PluginManager.get_plugin("cook_persistence_plug_point")

    _restaurant_persistence.print_all()
    manager_name = "Vinny Gambini"
    manager_email = "vinny@unreal.com"
    manager_password = "secret"
    restaurant_name = "Siciliano Ristorante"
    restaurant_address = "1 Pizza Way"
    restaurant_cuisine_type = "Italian"
    cooks = {
        "giada": {"name": "Giada De Laurentiis", "email": "gdl@unreal.com", "password": "secret"},
        "fabio": {"name": "Fabio Viviani", "email": "fabio@unreal.com", "password": "secret"}
    }
    result, message = rats.signup_initiate(manager_name, manager_email, restaurant_name)
    assert result == "success"
    assert message == "restaurant signup initiated"

    verification_code = rats._signup_get_verification_code(manager_email, restaurant_name)
    result, message = rats.signup_verify(manager_email, restaurant_name, verification_code)
    assert result == "success"
    assert message == "restaurant signup verified"

    result, message = rats.signup_complete(restaurant_name, restaurant_address, restaurant_cuisine_type)
    assert result == "success"
    assert message == "restaurant signed-up"

    rats.signup(manager_email, manager_password)

    manager_secure_token = rats.signin(manager_email, manager_password)
    cook_g = rats.add_cook(manager_secure_token, cooks["giada"]["name"], cooks["giada"]["email"])
    cook_f = rats.add_cook(manager_secure_token, cooks["fabio"]["name"], cooks["fabio"]["email"])
    assert _cook_persistence.read_cook_count() == 2

    rats.remove_cook(manager_secure_token, cook_g)
    assert _cook_persistence.read_cook_count() == 1

    assert _restaurant_persistence.read_restaurant_count() == 1

    _restaurant_persistence.print_all()

    """WIP these are not fully built but as you build them update the test cases"""
    rats.upload_menu()

    """OPROC cook giada signs in """
    result, message = oproc.signup(cooks["giada"]["email"], cooks["giada"]["password"])
    assert result == "success"
    assert message == "signed-up actor='cook', email='gdl@unreal.com'"

    """ 
    OPROC cook giada see if there is any open orders she needs to work on.
    and since no customer has placed an order, the queue shows empty.
    """

    """
    OCRA: customer Daren Daring signs in
    """
    customer_name = "Daren Daring"
    customer_email = "dd@unreal.com"
    customer_password = "password"
    result, message = ocra.signup(customer_email, customer_password)
    customer_secure_token = ocra.signin(customer_email, customer_password)
    assert customer_secure_token is not None

    customer_orders = ocra.read_orders(customer_secure_token)
    assert len(customer_orders) == 0
    """
    OCRA: customer Daren selects one item, and does checkout
    no payment, nothing. just simple checkout
    """
    result, message = ocra.checkout(customer_secure_token, "cheese pizza")
    assert result == "success"
    assert message == ""
    print(f'{datetime.now()}:ocra.checkout {result=}')

    customer_orders = ocra.read_orders(customer_secure_token)
    assert len(customer_orders) == 1
    print(f'{datetime.now()}:ocra.read_orders {len(customer_orders)=}')

    """give eventing a chance to get caught up"""
    print(f'{datetime.now()}:sleep for {PAUSE_PERIOD} seconds')
    await asyncio.sleep(PAUSE_PERIOD)

    print(f'{datetime.now()}:OPROC starting now. cook wants to the orders')
    """OPROC: cook giada now sees there is an order pending"""
    cook_secure_token = oproc.signin(cooks["giada"]["email"], cooks["giada"]["password"])
    assert cook_secure_token is not None
    orders = oproc.view_open_order_queue(cook_secure_token)
    assert len(orders) == 1

    rats.view_order_status()
    rats.restaurant_rated()
    print(f'{datetime.now()}:done with prepared test steps')
    print(f'{datetime.now()}:===============================================')


# NOTE: this test needs to be the second test as it expects previous method to initialize
@pytest.mark.asyncio
async def test_signup_happy_path_cuban():
    _restaurant_persistence.print_all()
    manager_name = "Pico Martinez"
    manager_email = "pmartinez@unreal.com"
    restaurant_name = "Cuba Casa"
    restaurant_address = "123 Havana Ave"
    restaurant_cuisine_type = "Cuban"
    result, message = rats.signup_initiate(manager_name, manager_email, restaurant_name)
    assert result == "success"
    assert message == "restaurant signup initiated"

    verification_code = rats._signup_get_verification_code(manager_email, restaurant_name)
    result, message = rats.signup_verify(manager_email, restaurant_name, verification_code)
    assert result == "success"
    assert message == "restaurant signup verified"

    result, message = rats.signup_complete(restaurant_name, restaurant_address, restaurant_cuisine_type)
    assert result == "success"
    assert message == "restaurant signed-up"

    # make test independent. right now it is dependent on first one
    assert _restaurant_persistence.read_restaurant_count() == 2
    _restaurant_persistence.print_all()


@pytest.fixture()
def setup_all_persistence():
    # making sure the test always starts with empty list
    _restaurant_persistence.RESTAURANT_DICT = {}
    _cook_persistence.clear_all()
    _signin_persistence.clear_all()
    _signup_persistence.clear_all()
    yield
    # making sure the test cleans up to an empty list
    _restaurant_persistence.RESTAURANT_DICT = {}
    _cook_persistence.COOK_DICT = {}
    _signin_persistence.clear_all()
    _signup_persistence.clear_all()


@pytest.mark.asyncio
async def test_signup_error_paths(setup_all_persistence):
    """
    in this path we will go through most of the steps but by going through error paths
    """
    core_main.initialize_app()

    _restaurant_persistence.print_all()
    manager_name = "Errico Errorprone"
    manager_email = "ee@unreal.com"
    manager_password = "secret"
    restaurant_name = "Earthy Flavors"
    restaurant_address = "N Delicious Way"
    restaurant_cuisine_type = "Fusion"
    typo_error = restaurant_name + "bad_string"

    cooks = {
        "sb": {"name": "sb", "email": "sb@unreal.com"}
    }
    result, message = rats.signup_initiate(manager_name, manager_email, restaurant_name)
    assert result == "success"
    assert message == "restaurant signup initiated"

    """
    Errico forgets he signed up, and he signs up again
    this time, we catch it and tell him the restaurant is already signed up
    """
    result, message = rats.signup_initiate(manager_name, manager_email, restaurant_name)
    assert result == "error"
    assert message == "restaurant already signed up"

    """
    user realizes his folly. he remembers he had started the process
    and resumes the verification process.
    """
    result, message = rats.signup_verify(manager_email, typo_error, "dont_know")
    assert result == "error"
    assert message == f'invalid restaurant_name=\'{typo_error}\''
    """
    he types in restaurant name wrong. he fixes that.
    BUT, he doesn't remember the verification code. he just guesses it
    """
    result, message = rats.signup_verify(manager_email, restaurant_name, "dont_know")
    assert result == "error"
    assert message == f'invalid key'

    """
    he now gets serious.
    he looks in his email inbox for the verification code
    he finds it and enters it
    """
    verification_code = rats._signup_get_verification_code(manager_email, restaurant_name)
    result, message = rats.signup_verify(manager_email, restaurant_name, verification_code)
    assert result == "success"
    assert message == "restaurant signup verified"

    """
    almost done with signup but he types incorrect restaurant name 
    """
    result, message = rats.signup_complete(typo_error, restaurant_address, restaurant_cuisine_type)
    assert result == "error"
    assert message == f'invalid restaurant_name=\'{typo_error}\''

    result, message = rats.signup_complete(restaurant_name, restaurant_address, restaurant_cuisine_type)
    assert result == "success"
    assert message == "restaurant signed-up"

    rats.signup(manager_email, manager_password)

    manager_secure_token = rats.signin(manager_email, manager_password)
    cook_g = rats.add_cook(manager_secure_token, cooks["sb"]["name"], cooks["sb"]["email"])
    assert _cook_persistence.read_cook_count() == 1

    rats.remove_cook(manager_secure_token, cook_g)
    assert _cook_persistence.read_cook_count() == 0


def test_empty():
    event_bus_default = EventBusInterface()
    event_bus_default.add_listener("dummy", None)
    event_bus_default.emit("dummy", {})



"""
ocra.py
this follows interactor pattern.
implements both command executor and event listener

see common notes in rats.py
"""
from qb_core.ocra_pkg.order import order_entity
from qb_core.ocra_pkg.order.order_model import Order, OrderItem, OrderAdded
from qb_core.rats_pkg.signin import signin_service
from qb_core.event_bus import event_bus_core, event_bus_events


def initialize_module():
    event_bus_core.register_emitter(event_bus_events.EVENT_ORDER_CONFIRMED, None)
    event_bus_core.register_emitter(event_bus_events.EVENT_RESTAURANT_RATED, None)
    event_bus_core.register_emitter(event_bus_events.EVENT_DRIVER_RATED, None)

    print(f'*** REGISTERING listener None added for event menu_uploaded')
    event_bus_core.register_listener(event_bus_events.EVENT_MENU_UPLOADED, None)


def signup(customer_email, password):
    return signin_service.signup(signin_service.ACTOR_CUSTOMER, customer_email, password)


def signin(customer_email, password):
    return signin_service.signin(signin_service.ACTOR_CUSTOMER, customer_email, password)


def checkout(secure_token, order_item_name):
    order = Order()
    # TODO how to generate this number
    order.id = "123123"
    order.customer_email = secure_token.email
    order_item = OrderItem()
    order_item.item_name = order_item_name
    order_item.quantity = 1
    order_item.price = 9.99
    order.items.append(order_item)
    order_entity.save_order(order)

    # this should be the last step so that we send this event exactly once
    # when all the steps above have completed
    event_bus_core.emit_event(event_bus_events.EVENT_ORDER_CONFIRMED, OrderAdded(order))

    return "success", ""


def read_orders(secure_token):
    return order_entity.read_orders(secure_token.email)



CUSTOMER_DICT = {}


def read_customer_count():
    return len(CUSTOMER_DICT)


def read_customer(customer_name):
    return CUSTOMER_DICT.get(customer_name)


def save_customer(customer):
    CUSTOMER_DICT[customer.email] = customer


def remove_customer(customer):
    CUSTOMER_DICT.pop(customer.email)


def print_all():
    print(f'{CUSTOMER_DICT=}')



SIGNUP_DICT = {}


def clear_all():
    SIGNIN_DICT = {}
    SIGNIN_SESSION = {}


def read_code(key):
    return SIGNUP_DICT.get(key)


def save_code(key, code):
    SIGNUP_DICT[key] = code


# hey, lets not get into the habit of printing sensitive info
# so turn this off
# WARNING never turn this ON
# def print_all():
#     print(f'{SIGNUP_DICT=}')



from collections import defaultdict

from qb_core.rats_pkg.cook.cook_persistence_plug_point import CookPersistenceInterface

cook_dict = {}


class CookPersistenceCore(CookPersistenceInterface):
    def __init__(self):
        pass

    @staticmethod
    def clear_all():
        global cook_dict
        cook_dict = defaultdict(dict)

    def read_cook_count(self):
        return len(cook_dict)

    def read_cook(self, cook_name):
        return cook_dict.get(cook_name)

    def save_cook(self, cook):
        cook_dict[cook.name] = cook

    def remove_cook(self, cook):
        cook_dict.pop(cook.name)

    def print_all(self):
        print(f'{cook_dict=}')


def get_implementation():
    return CookPersistenceCore()


from collections import defaultdict

ORDER_LIST = []


def clear_all():
    global ORDER_LIST
    ORDER_LIST = []


def read_order_count(customer_email):
    return len(ORDER_LIST[customer_email])


def read_open_orders():
    open_orders = []
    for o in ORDER_LIST:
        # open_orders.append(o)
        if o.cook_assignment is None:
            open_orders.append(o)
    return open_orders


def save_order(order):
    print(f'OPROC order_persistence.save_order called')
    ORDER_LIST.append(order)
    print_all()
    print(f'OPROC order_persistence.save_order returns')


def print_all():
    print(f'{ORDER_LIST=}')


clear_all()



import asyncio
from collections import defaultdict

from qb_core.event_bus.event_bus_interface import EventBusInterface


class EventBusDefault(EventBusInterface):
    def __init__(self):
        self.listeners = defaultdict(set)

    def add_listener(self, event_name, listener):
        self.listeners[event_name].add(listener)

    # I don't think I need this
    # def remove_listener(self, event_name, listener):
    #     self.listeners[event_name].remove(listener)
    #     if len(self.listeners[event_name]) == 0:
    #         del self.listeners[event_name]

    def emit(self, event_name, event):
        listeners = self.listeners.get(event_name, [])
        print(f'emitting {event_name=} with {str(event)=} to {len(listeners)} listeners')
        for listener in listeners:
            print(f'{event_name=} with {event=} sent to {str(listener)}')
            asyncio.create_task(listener(event))
            # # asyncio suggests this pattern
            # # https://docs.python.org/3/library/asyncio-task.html#creating-tasks
            # Save a reference to the result of this function, to avoid a task disappearing mid execution.
            # The event loop only keeps weak references to tasks.
            # A task that isn’t referenced elsewhere may get garbage-collected at any time, even before it’s done.
            # For reliable “fire-and-forget” background tasks, gather them in a collection
            # task = asyncio.create_task(some_coro(param=i))
            # # Add task to the set. This creates a strong reference.
            # background_tasks.add(task)
            #
            # # To prevent keeping references to finished tasks forever,
            # # make each task remove its own reference from the set after
            # # completion:
            # task.add_done_callback(background_tasks.discard)



"""cook_entity.py"""
from qb_core.common.plugin.plugin_manager import PluginManager

"""
this provides cook service
basically all the functions related to cook are included here.
this is the pluggable class that invokes plug points 
"""


def add_cook(cook):
    # cook specific validations go here
    # example, are all fields correct?
    _get_cook_plugin().save_cook(cook)


def remove_cook(cook):
    # cook specific validations go here
    _get_cook_plugin().remove_cook(cook)


def _get_cook_plugin():
    """
    convinence method to get the plugin for this plug point
    :return:
    """
    # tip: don't save this in this class. when core initializes/imports this module
    # infrastructure code would not have had opportunity to override. this why it is
    # important to always go to plugin manager for the most current plugin
    return PluginManager.get_plugin("cook_persistence_plug_point")



from qb_core.ocra_pkg.customer import _customer_persistence


def add_customer(customer):
    _customer_persistence.save_customer(customer)


def remove_customer(customer):
    _customer_persistence.remove_customer(customer)



"""
oproc.py
this is Order Processor
"""
from datetime import datetime

from qb_core.event_bus import event_bus_core, event_bus_events
from qb_core.oproc_pkg.order import order_entity
from qb_core.oproc_pkg.order.order_model import Order
from qb_core.rats_pkg.signin import signin_service


def initialize_module():
    # this must be called exactly once, so put it here at the module level
    # can a module be imported more than once?
    print(f'*** REGISTERING listener handle_cooked added for event cook_added')
    event_bus_core.register_listener(event_bus_events.EVENT_COOK_ADDED, handle_cook_added)
    event_bus_core.register_listener(event_bus_events.EVENT_ORDER_CONFIRMED, handle_order_confirmed)


async def handle_cook_added(event_data):
    print(f'handle_cook_added called with {event_data=}')


async def handle_order_confirmed(event_data):
    try:
        print(
            f'{datetime.now()}:handle_order_confirmed called with {event_data=}. got order of type {type(event_data.order)}')
        oproc_order = Order()
        oproc_order.id = event_data.order.id
        oproc_order.customer_email = event_data.order.customer_email
        oproc_order.total_price = event_data.order.total_price
        oproc_order.items = event_data.order.items
        order_entity.save_order(oproc_order)
        print(f'{datetime.now()}:done saving order {order_entity.read_open_orders()=}')
        order_entity.read_open_orders()
    except Exception as ex:
        print(f'{datetime.now()}: caught {str(ex)}')


def signup(cook_email, password):
    return signin_service.signup(signin_service.ACTOR_COOK, cook_email, password)


def signin(cook_email, password):
    return signin_service.signin(signin_service.ACTOR_COOK, cook_email, password)


def view_open_order_queue(cook_secure_token):
    return order_entity.read_open_orders()



import asyncio
from collections import defaultdict


class EventBusInterface:
    def __init__(self):
        pass

    def add_listener(self, event_name, listener):
        pass

    # I don't think I need this
    # def remove_listener(self, event_name, listener):
    #     pass

    def emit(self, event_name, event):
        pass



"""
basic things I need in an event bus
* emitter registers with EB it's intent emit an event
* it describes the event object
* one or more listener register with EB their intent to subscribe to an event
* emitter emits an event to EB
* EB invokes the listeners registered

this class acts as interface class
"""
from qb_core.event_bus.event_bus_default import EventBusDefault

# TODO make this conditional
#  if plugin is not configured (or configured to default) do this, else use plugin suggested event bus
system_event_bus = EventBusDefault()
print(f'*** initialized {system_event_bus=}')


def register_emitter(event_name, emitter_name):
    """
    use this method for registering the emitter
    """
    pass


def register_listener(event_name, listener):
    system_event_bus.add_listener(event_name, listener)


def emit_event(event_name, event):
    system_event_bus.emit(event_name, event)

